#ifndef OSU_H
#define OSU_H

#include "mpi.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/time.h>

#define FIELD_WIDTH 20
#define FLOAT_PRECISION 2
#define OMB_VERSION "v3.1.1"

#endif
